#!/usr/bin/env python

"""
Web interface data abstractions.

Copyright (C) 2014, 2015, 2017 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime, timedelta
from imiptools.dates import end_date_from_calendar, end_date_to_calendar, \
                            format_datetime, get_datetime, get_end_of_day, \
                            to_date
from imiptools.period import RecurringPeriod

class PeriodError(Exception):
    pass

class EventPeriod(RecurringPeriod):

    """
    A simple period plus attribute details, compatible with RecurringPeriod, and
    intended to represent information obtained from an iCalendar resource.
    """

    def __init__(self, start, end, tzid=None, origin=None, start_attr=None,
                 end_attr=None, form_start=None, form_end=None, replaced=False):

        """
        Initialise a period with the given 'start' and 'end' datetimes, together
        with optional 'start_attr' and 'end_attr' metadata, 'form_start' and
        'form_end' values provided as textual input, and with an optional
        'origin' indicating the kind of period this object describes.
        """

        RecurringPeriod.__init__(self, start, end, tzid, origin, start_attr, end_attr)
        self.form_start = form_start
        self.form_end = form_end
        self.replaced = replaced

    def as_tuple(self):
        return self.start, self.end, self.tzid, self.origin, self.start_attr, \
               self.end_attr, self.form_start, self.form_end, self.replaced

    def __repr__(self):
        return "EventPeriod%r" % (self.as_tuple(),)

    def as_event_period(self):
        return self

    def get_start_item(self):
        return self.get_start(), self.get_start_attr()

    def get_end_item(self):
        return self.get_end(), self.get_end_attr()

    # Form data compatibility methods.

    def get_form_start(self):
        if not self.form_start:
            self.form_start = self.get_form_date(self.get_start(), self.start_attr)
        return self.form_start

    def get_form_end(self):
        if not self.form_end:
            self.form_end = self.get_form_date(end_date_from_calendar(self.get_end()), self.end_attr)
        return self.form_end

    def as_form_period(self):
        return FormPeriod(
            self.get_form_start(),
            self.get_form_end(),
            isinstance(self.end, datetime) or self.get_start() != self.get_end() - timedelta(1),
            isinstance(self.start, datetime) or isinstance(self.end, datetime),
            self.tzid,
            self.origin,
            self.replaced and True or False
            )

    def get_form_date(self, dt, attr=None):
        return FormDate(
            format_datetime(to_date(dt)),
            isinstance(dt, datetime) and str(dt.hour) or None,
            isinstance(dt, datetime) and str(dt.minute) or None,
            isinstance(dt, datetime) and str(dt.second) or None,
            attr and attr.get("TZID") or None,
            dt, attr
            )

class FormPeriod(RecurringPeriod):

    "A period whose information originates from a form."

    def __init__(self, start, end, end_enabled=True, times_enabled=True,
                 tzid=None, origin=None, replaced=False):
        self.start = start
        self.end = end
        self.end_enabled = end_enabled
        self.times_enabled = times_enabled
        self.tzid = tzid
        self.origin = origin
        self.replaced = replaced

    def as_tuple(self):
        return self.start, self.end, self.end_enabled, self.times_enabled, self.tzid, self.origin, self.replaced

    def __repr__(self):
        return "FormPeriod%r" % (self.as_tuple(),)

    def __cmp__(self, other):
        result = RecurringPeriod.__cmp__(self, other)
        if result:
            return result
        other = form_period_from_period(other)
        return cmp(self.replaced, other.replaced)

    def as_event_period(self, index=None):

        """
        Return a converted version of this object as an event period suitable
        for iCalendar usage. If 'index' is indicated, include it in any error
        raised in the conversion process.
        """

        dtstart, dtstart_attr = self.get_start_item()
        if not dtstart:
            if index is not None:
                raise PeriodError(("dtstart", index))
            else:
                raise PeriodError("dtstart")

        dtend, dtend_attr = self.get_end_item()
        if not dtend:
            if index is not None:
                raise PeriodError(("dtend", index))
            else:
                raise PeriodError("dtend")

        if dtstart > dtend:
            if index is not None:
                raise PeriodError(("dtstart", index), ("dtend", index))
            else:
                raise PeriodError("dtstart", "dtend")

        return EventPeriod(dtstart, end_date_to_calendar(dtend), self.tzid,
                           self.origin, dtstart_attr, dtend_attr,
                           self.start, self.end, self.replaced)

    # Period data methods.

    def get_start(self):
        return self.start and self.start.as_datetime(self.times_enabled) or None

    def get_end(self):

        # Handle specified end datetimes.

        if self.end_enabled:
            dtend = self.end.as_datetime(self.times_enabled)
            if not dtend:
                return None

        # Handle same day times.

        elif self.times_enabled:
            formdate = FormDate(self.start.date, self.end.hour, self.end.minute, self.end.second, self.end.tzid)
            dtend = formdate.as_datetime(self.times_enabled)
            if not dtend:
                return None

        # Otherwise, treat the end date as the start date. Datetimes are
        # handled by making the event occupy the rest of the day.

        else:
            dtstart, dtstart_attr = self.get_start_item()
            if dtstart:
                if isinstance(dtstart, datetime):
                    dtend = get_end_of_day(dtstart, dtstart_attr["TZID"])
                else:
                    dtend = dtstart
            else:
                return None

        return dtend

    def get_start_attr(self):
        return self.start and self.start.get_attributes(self.times_enabled) or {}

    def get_end_attr(self):
        return self.end and self.end.get_attributes(self.times_enabled) or {}

    # Form data methods.

    def get_form_start(self):
        return self.start

    def get_form_end(self):
        return self.end

    def as_form_period(self):
        return self

class FormDate:

    "Date information originating from form information."

    def __init__(self, date=None, hour=None, minute=None, second=None, tzid=None, dt=None, attr=None):
        self.date = date
        self.hour = hour
        self.minute = minute
        self.second = second
        self.tzid = tzid
        self.dt = dt
        self.attr = attr

    def as_tuple(self):
        return self.date, self.hour, self.minute, self.second, self.tzid, self.dt, self.attr

    def __repr__(self):
        return "FormDate%r" % (self.as_tuple(),)

    def get_component(self, value):
        return (value or "").rjust(2, "0")[:2]

    def get_hour(self):
        return self.get_component(self.hour)

    def get_minute(self):
        return self.get_component(self.minute)

    def get_second(self):
        return self.get_component(self.second)

    def get_date_string(self):
        return self.date or ""

    def get_datetime_string(self):
        if not self.date:
            return ""

        hour = self.hour; minute = self.minute; second = self.second

        if hour or minute or second:
            time = "T%s%s%s" % tuple(map(self.get_component, (hour, minute, second)))
        else:
            time = ""
            
        return "%s%s" % (self.date, time)

    def get_tzid(self):
        return self.tzid

    def as_datetime(self, with_time=True):

        "Return a datetime for this object."

        # Return any original datetime details.

        if self.dt:
            return self.dt

        # Otherwise, construct a datetime.

        s, attr = self.as_datetime_item(with_time)
        if s:
            return get_datetime(s, attr)
        else:
            return None

    def as_datetime_item(self, with_time=True):

        """
        Return a (datetime string, attr) tuple for the datetime information
        provided by this object, where both tuple elements will be None if no
        suitable date or datetime information exists.
        """

        s = None
        if with_time:
            s = self.get_datetime_string()
            attr = self.get_attributes(True)
        if not s:
            s = self.get_date_string()
            attr = self.get_attributes(False)
        if not s:
            return None, None
        return s, attr

    def get_attributes(self, with_time=True):

        "Return attributes for the date or datetime represented by this object."

        if with_time:
            return {"TZID" : self.get_tzid(), "VALUE" : "DATE-TIME"}
        else:
            return {"VALUE" : "DATE"}

def event_period_from_period(period):

    """
    Convert a 'period' to one suitable for use in an iCalendar representation.
    In an "event period" representation, the end day of any date-level event is
    encoded as the "day after" the last day actually involved in the event.
    """

    if isinstance(period, EventPeriod):
        return period
    elif isinstance(period, FormPeriod):
        return period.as_event_period()
    else:
        dtstart, dtstart_attr = period.get_start_item()
        dtend, dtend_attr = period.get_end_item()
        if not isinstance(period, RecurringPeriod):
            dtend = end_date_to_calendar(dtend)
        return EventPeriod(dtstart, dtend, period.tzid, period.origin, dtstart_attr, dtend_attr)

def form_period_from_period(period):

    """
    Convert a 'period' into a representation usable in a user-editable form.
    In a "form period" representation, the end day of any date-level event is
    presented in a "natural" form, not the iCalendar "day after" form.
    """

    if isinstance(period, EventPeriod):
        return period.as_form_period()
    elif isinstance(period, FormPeriod):
        return period
    else:
        return event_period_from_period(period).as_form_period()



# Form field extraction and serialisation.

def get_date_control_inputs(args, name, tzid_name=None):

    """
    Return a tuple of date control inputs taken from 'args' for field names
    starting with 'name'.

    If 'tzid_name' is specified, the time zone information will be acquired
    from fields starting with 'tzid_name' instead of 'name'.
    """

    return args.get("%s-date" % name, []), \
           args.get("%s-hour" % name, []), \
           args.get("%s-minute" % name, []), \
           args.get("%s-second" % name, []), \
           args.get("%s-tzid" % (tzid_name or name), [])

def get_date_control_values(args, name, multiple=False, tzid_name=None, tzid=None):

    """
    Return a form date object representing fields taken from 'args' starting
    with 'name'.

    If 'multiple' is set to a true value, many date objects will be returned
    corresponding to a collection of datetimes.

    If 'tzid_name' is specified, the time zone information will be acquired
    from fields starting with 'tzid_name' instead of 'name'.

    If 'tzid' is specified, it will provide the time zone where no explicit
    time zone information is indicated in the field data.
    """

    dates, hours, minutes, seconds, tzids = get_date_control_inputs(args, name, tzid_name)

    # Handle absent values by employing None values.

    field_values = map(None, dates, hours, minutes, seconds, tzids)

    if not field_values and not multiple:
        all_values = FormDate()
    else:
        all_values = []
        for date, hour, minute, second, tzid_field in field_values:
            value = FormDate(date, hour, minute, second, tzid_field or tzid)

            # Return a single value or append to a collection of all values.

            if not multiple:
                return value
            else:
                all_values.append(value)

    return all_values

def set_date_control_values(formdates, args, name, tzid_name=None):

    """
    Using the values of the given 'formdates', replace form fields in 'args'
    starting with 'name'.

    If 'tzid_name' is specified, the time zone information will be stored in
    fields starting with 'tzid_name' instead of 'name'.
    """

    args["%s-date" % name] = []
    args["%s-hour" % name] = []
    args["%s-minute" % name] = []
    args["%s-second" % name] = []
    args["%s-tzid" % (tzid_name or name)] = []

    for d in formdates:
        args["%s-date" % name].append(d and d.date or "")
        args["%s-hour" % name].append(d and d.hour or "")
        args["%s-minute" % name].append(d and d.minute or "")
        args["%s-second" % name].append(d and d.second or "")
        args["%s-tzid" % (tzid_name or name)].append(d and d.tzid or "")

def get_period_control_values(args, start_name, end_name,
                              end_enabled_name, times_enabled_name,
                              origin=None, origin_name=None,
                              replaced_name=None, tzid=None):

    """
    Return period values from fields found in 'args' prefixed with the given
    'start_name' (for start dates), 'end_name' (for end dates),
    'end_enabled_name' (to enable end dates for periods), 'times_enabled_name'
    (to enable times for periods).

    If 'origin' is specified, a single period with the given origin is
    returned. If 'origin_name' is specified, fields containing the name will
    provide origin information, and fields containing 'replaced_name' will
    indicate periods that are replaced.

    If 'tzid' is specified, it will provide the time zone where no explicit
    time zone information is indicated in the field data.
    """

    # Get the end datetime and time presence settings.

    all_end_enabled = args.get(end_enabled_name, [])
    all_times_enabled = args.get(times_enabled_name, [])

    # Get the origins of period data and whether the periods are replaced.

    if origin:
        all_origins = [origin]
    else:
        all_origins = origin_name and args.get(origin_name, []) or []

    all_replaced = replaced_name and args.get(replaced_name, []) or []

    # Get the start and end datetimes.

    all_starts = get_date_control_values(args, start_name, True, tzid=tzid)
    all_ends = get_date_control_values(args, end_name, True, start_name, tzid=tzid)

    # Construct period objects for each start, end, origin combination.

    periods = []

    for index, (start, end, found_origin) in \
        enumerate(map(None, all_starts, all_ends, all_origins)):

        # Obtain period settings from separate controls.

        end_enabled = str(index) in all_end_enabled
        times_enabled = str(index) in all_times_enabled
        replaced = str(index) in all_replaced

        period = FormPeriod(start, end, end_enabled, times_enabled, tzid,
                            found_origin or origin, replaced)
        periods.append(period)

    # Return a single period if a single origin was specified.

    if origin:
        return periods[0]
    else:
        return periods

def set_period_control_values(periods, args, start_name, end_name,
                              end_enabled_name, times_enabled_name,
                              origin_name=None, replaced_name=None):

    """
    Using the given 'periods', replace form fields in 'args' prefixed with the
    given 'start_name' (for start dates), 'end_name' (for end dates),
    'end_enabled_name' (to enable end dates for periods), 'times_enabled_name'
    (to enable times for periods).

    If 'origin_name' is specified, fields containing the name will provide
    origin information, and fields containing 'replaced_name' will indicate
    periods that are replaced.
    """

    # Record period settings separately.

    args[end_enabled_name] = []
    args[times_enabled_name] = []

    # Record origin and replacement information if naming is defined.

    if origin_name:
        args[origin_name] = []

    if replaced_name:
        args[replaced_name] = []

    all_starts = []
    all_ends = []

    for index, period in enumerate(periods):

        # Encode period settings in controls.

        if period.end_enabled:
            args[end_enabled_name].append(str(index))
        if period.times_enabled:
            args[times_enabled_name].append(str(index))

        # Add origin information where controls are present to record it.

        if origin_name:
            args[origin_name].append(period.origin or "")

        # Add replacement information where controls are present to record it.

        if replaced_name and period.replaced:
            args[replaced_name].append(str(index))

        # Collect form date information for addition below.

        all_starts.append(period.get_form_start())
        all_ends.append(period.get_form_end())

    # Set the controls for the dates.

    set_date_control_values(all_starts, args, start_name)
    set_date_control_values(all_ends, args, end_name, tzid_name=start_name)

# vim: tabstop=4 expandtab shiftwidth=4
